from dune.perftool.generation import (get_global_context_value,
                                      include_file,
                                      preamble,
                                      )
from dune.perftool.pdelab.driver import (form_name_suffix,
                                         get_cell,
                                         get_dimension,
                                         get_test_element,
                                         get_trial_element,
                                         isQuadrilateral,
                                         name_initree,
                                         )
from dune.perftool.pdelab.driver.constraints import (name_assembled_constraints,
                                                     type_constraintscontainer,
                                                     )
from dune.perftool.pdelab.driver.gridfunctionspace import (name_test_gfs,
                                                           name_trial_gfs,
                                                           preprocess_leaf_data,
                                                           type_domainfield,
                                                           type_range,
                                                           type_test_gfs,
                                                           type_trial_gfs,
                                                           )
from dune.perftool.pdelab.localoperator import localoperator_basename
from dune.perftool.pdelab.parameter import parameterclass_basename
from dune.perftool.options import get_option


@preamble
def typedef_gridoperator(name, formdata):
    ugfs = type_trial_gfs()
    vgfs = type_test_gfs()
    lop = type_localoperator(formdata)
    cc = type_constraintscontainer()
    mb = type_matrixbackend()
    df = type_domainfield()
    r = type_range()
    if get_option("fastdg"):
        if not get_option("sumfact"):
            raise PerftoolCodegenError("FastDGGridOperator is only implemented for sumfactorization.")
        include_file("dune/pdelab/gridoperator/fastdg.hh", filetag="driver")
        return "using {} = Dune::PDELab::FastDGGridOperator<{}, {}, {}, {}, {}, {}, {}, {}, {}>;".format(name, ugfs, vgfs, lop, mb, df, r, r, cc, cc)
    else:
        include_file("dune/pdelab/gridoperator/gridoperator.hh", filetag="driver")
        return "using {} = Dune::PDELab::GridOperator<{}, {}, {}, {}, {}, {}, {}, {}, {}>;".format(name, ugfs, vgfs, lop, mb, df, r, r, cc, cc)


def type_gridoperator(formdata):
    name = form_name_suffix("GO", formdata).upper()
    typedef_gridoperator(name, formdata)
    return name


@preamble
def define_gridoperator(name, formdata):
    gotype = type_gridoperator(formdata)
    ugfs = name_trial_gfs()
    vgfs = name_test_gfs()
    if ugfs != vgfs:
        raise NotImplementedError("Non-Galerkin methods currently not supported!")
    cc = name_assembled_constraints()
    lop = name_localoperator(formdata)
    mb = name_matrixbackend()
    return ["{} {}({}, {}, {}, {}, {}, {});".format(gotype, name, ugfs, cc, vgfs, cc, lop, mb),
            "std::cout << \"gfs with \" << {}.size() << \" dofs generated  \"<< std::endl;".format(ugfs),
            "std::cout << \"cc with \" << {}.size() << \" dofs generated  \"<< std::endl;".format(cc)]


def name_gridoperator(formdata):
    name = form_name_suffix("go", formdata)
    define_gridoperator(name, formdata)
    return name


@preamble
def typedef_localoperator(name, formdata):
    ugfs = type_trial_gfs()
    vgfs = type_test_gfs()
    data = get_global_context_value("data")
    filename = get_option("operator_file")
    include_file(filename, filetag="driver")
    lopname = localoperator_basename(formdata, data)
    range_type = type_range()
    return "using {} = {}<{}, {}, {}>;".format(name, lopname, ugfs, vgfs, range_type)


def type_localoperator(formdata):
    name = form_name_suffix("LOP", formdata).upper()
    typedef_localoperator(name, formdata)
    return name


@preamble
def define_localoperator(name, formdata):
    loptype = type_localoperator(formdata)
    ini = name_initree()
    params = name_parameters(formdata)
    return "{} {}({}, {});".format(loptype, name, ini, params)


def name_localoperator(formdata):
    name = form_name_suffix("lop", formdata)
    define_localoperator(name, formdata)
    return name


@preamble
def define_dofestimate(name):
    # Provide a worstcase estimate for the number of entries per row based
    # on the given gridfunction space and cell geometry
    if isQuadrilateral(get_cell()):
        geo_factor = 2**get_dimension()
    else:
        if get_dimension() < 3:
            geo_factor = 3 * get_dimension()
        else:
            # TODO no idea how a generic estimate for 3D simplices looks like
            geo_factor = 12

    gfs = name_trial_gfs()
    ini = name_initree()

    return ["{}.update();".format(gfs),
            "int generic_dof_estimate =  {} * {}.maxLocalSize();".format(geo_factor, gfs),
            "int {} = {}.get<int>(\"istl.number_of_nnz\", generic_dof_estimate);".format(name, ini)]


def name_dofestimate():
    name = "dofestimate"
    define_dofestimate(name)
    return name


@preamble
def typedef_matrixbackend(name):
    include_file("dune/pdelab/backend/istl.hh", filetag="driver")
    return "using {} = Dune::PDELab::ISTL::BCRSMatrixBackend<>;".format(name)


def type_matrixbackend():
    name = "MatrixBackend"
    typedef_matrixbackend(name)
    return name


@preamble
def define_matrixbackend(name):
    mbtype = type_matrixbackend()
    dof = name_dofestimate()
    return "{} {}({});".format(mbtype, name, dof)


def name_matrixbackend():
    name = "mb"
    define_matrixbackend(name)
    return name


def type_parameters(formdata):
    data = get_global_context_value("data")
    name = parameterclass_basename(formdata, data)
    return name


@preamble
def define_parameters(name, formdata):
    partype = type_parameters(formdata)
    return "{} {};".format(partype, name)


def name_parameters(formdata):
    name = form_name_suffix("params", formdata)
    define_parameters(name, formdata)
    return name
